/*
 * $Id: net.c,v 1.4 2013/03/01 19:16:08 jacadcaps Exp $
 */

// #define DEBUG

#include "globals.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <amitcp/socketbasetags.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/filio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <proto/keymap.h>
#include <proto/socket.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <exec/rawfmt.h>

#include "rbf.h"
#include "des.h"
#include "errorreq.h"
#include "net.h"
#include "args.h"


#define REPORT_RECV_FAILURES 0 /* set to have a feedback on every recv() failure */
#define REPORT_SEND_FAILURES 0 /* ditto */

struct Library *SocketBase;
static int sock = -1;
static struct mouseevent me; /* global variable because InputEvents have deltamoves */
static struct keyevent ke;
static ULONG screen_width;
static ULONG screen_height;

#if REPORT_RECV_FAILURES
static int recv_fail;
#endif

#if REPORT_SEND_FAILURES
static int send_fail;
#endif

int tcp_earlyinit(void)
{
	if (SocketBase = OpenLibrary("bsdsocket.library", 3))
	{
		return 1;
	}
	return 0;
}

void tcp_latecleanup(void)
{
	if (SocketBase)
	{
		CloseLibrary(SocketBase);
	}
}

int tcp_connect(STRPTR host, int port)
{
	struct hostent *h;
	struct sockaddr_in sa;
	struct in_addr ip;
	struct timeval tv;
	int opt;

	if (SocketBase)
	{
		int err = 0; /* shut up gcc */
		long errno = 0; /* shut up gcc */
		ULONG i;

		ip.s_addr = inet_addr(host);

		if (ip.s_addr == INADDR_NONE)
		{
			if (h = gethostbyname(host))
			{
				DB(("connecting to %s\n", Inet_NtoA(*(u_long *)h->h_addr)));
				ip.s_addr = *(u_long *)h->h_addr;
			}
			else
			{
				errorreq(NULL, "Couldn't resolve host.", NULL);
				tcp_disconnect();
				return (FALSE);
			}
		}

		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (sock < 0)
		{   
			errorreq(NULL, "Unable to create socket.", NULL);
			tcp_disconnect();
			return (FALSE);
		}

		DB(("socketed\n"));

		memset(&sa, '\0', sizeof(struct sockaddr_in));
		sa.sin_family = AF_INET;
		sa.sin_port = port;
		sa.sin_addr.s_addr = ip.s_addr;

		tv.tv_sec = 16;
		tv.tv_usec = 0;
		setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (void *)&tv, sizeof(struct timeval));
		tv.tv_sec = 8;
		setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, (void *)&tv, sizeof(struct timeval));

		/* non blocking mode */
		opt = 1;
		IoctlSocket(sock, FIONBIO, (char *)&opt);

		/* wait 4-5 seconds */
		for (i = 0; i < 20; i++)
		{
			err = connect(sock, (struct sockaddr *)&sa, sizeof(struct sockaddr_in));
			errno = Errno();

			if (err < 0 && errno != EINPROGRESS && errno != EALREADY)
			{
				break;
			}
			Delay(10);
		}
		
		opt = 0;
		IoctlSocket(sock, FIONBIO, (char *)&opt);

		if (err < 0 && errno != EISCONN)
		{
			STRPTR errstr = (STRPTR)errno;
			ULONG l = 0;

			SocketBaseTags(SBTM_GETREF(SBTC_ERRNOSTRPTR), &errstr, TAG_DONE);

			NewRawDoFmt("Error while connect()ing:\n%s", (APTR)RAWFMTFUNC_COUNT, (STRPTR)&l, errstr);
			
			if (l > 0)
			{
				STRPTR errfmt = AllocVecTaskPooled(l + 1);
				
				if (errfmt)
				{
					NewRawDoFmt("Error while connect()ing:\n%s", (APTR)RAWFMTFUNC_STRING, errfmt, errstr);

					errorreq(NULL, errfmt, NULL);
			
					FreeVecTaskPooled(errfmt);
				}
			}

			DB(("errno == %ld\n", errno));
			tcp_disconnect();
			return (FALSE);
		}
		
		DB(("connected\n"));

		/*
		 * Set the right socket options.
		 */
		opt = 1;
		setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (void *)&opt, sizeof(opt));
		
		
		setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (void *)&opt, sizeof(opt));

		return (TRUE);
	}
	else
	{
		errorreq(NULL, "Needs bsdsocket.library 3 or higher.", NULL);
	}
	return (FALSE);
}


void tcp_disconnect(void)
{
	if (sock != -1)
	{
		CloseSocket(sock);
		sock = -1;
	}
}

LONG readn(int fd, APTR vptr, int n)
{
	int nleft, nread;
	char *ptr;

	ptr = vptr;
	nleft = n;

	while (nleft > 0)
	{
		if ((nread = recv(fd, ptr, nleft, 0)) < 0)
		{
			if (Errno() == EINTR)
			{
				nread = 0; /* restart */
			}
			else
			{
				return (-1);
			}
		}
		else if (nread == 0)
		{
			break;
		}

		nleft -= nread;
		ptr += nread;
	}
#if REPORT_RECV_FAILURES
	if (nleft)
	{
		DB(("recv() incomplete or failed\n"));
		if (!recv_fail++)
		{
			errorreq(NULL, "recv() incomplete or failed.", NULL);
		}
	}
#endif
	return (n - nleft);
}


/*
 * Not absolutely necessary since we block, but
 * who knows.. this is not guaranteed by some stacks
 */
LONG sendn(int fd, const APTR vptr, int n)
{
	int nleft, nwritten;
	const char *ptr;

	ptr = vptr;
	nleft = n;
	while (nleft > 0)
	{
		if ((nwritten = send(fd, ptr, nleft, 0)) <= 0)
		{
			if (Errno() == EINTR)
			{
				nwritten = 0; /* restart */
			}
			else
			{
				DB(("send() failed\n"));
#if REPORT_SEND_FAILURES
				if (!send_fail++)
				{
					errorreq(NULL, "send() failed.", NULL);
				}
#endif
				return (-1);
			}
		}
		nleft -= nwritten;
		ptr += nwritten;
	}
	return (n);
}


/*
 * ProtocolVersion 
 * Reads the server's version then
 * sends our version.
 */
int rbf_version(int version, int revision)
{
	char buf[14];
	if (readn(sock, buf, 12) == 12)
	{
		/*
		 * Consistency check. Are we really talking
		 * with a VNC server ?
		 */
		if (!strnicmp(buf, "RFB ", 4))
		{
			int minor, major;

			buf[7] = '\0';
			buf[11] = '\0';
			minor = atoi(&buf[4]);
			major = atoi(&buf[8]);
			DB(("major == %ld, minor == %ld\n", major, minor));
			
			/*
			 * We now send back our version supported.
			 */
			sprintf(buf, "RFB %03d.%03d\n", version, revision);
			if (sendn(sock, buf, 12) == 12)
			{
				return (TRUE);
			}
		}
		else
		{
			errorreq(NULL, "Remote doesn't appear to be a VNC server.", NULL);
		}
	}
	return (FALSE);
}


/*
 * Authentication
 */
int rbf_auth(STRPTR password)
{
	int r;

	if (readn(sock, &r, 4) == 4)
	{
		switch (r)
		{
			case 0: /* connection failed */
				rbf_display_connection_failed();
				break;
			
			case 1: /* no password needed */
				return (TRUE);

			case 2: /* password needed */
				if (rbf_handle_password(password))
				{
					return (TRUE);
				}
				break;

			default: /* unknown */
				errorreq(NULL, "Unknown authentication scheme.", "Quit");
				break;
		}
	}
	return (FALSE);
}	


/*
 * Display the reason of the connection failure.
 */
void rbf_display_connection_failed(void)
{
	ULONG size;	 

	if (readn(sock, &size, 4) == 4)
	{
		if (size < 4096) /* reasonable default */
		{
			STRPTR p;

			if (p = malloc(size + 1))
			{
				if (readn(sock, p, size) == size)
				{
					p[size + 1] = '\0';	
					errorreq(NULL, p, NULL);
					free(p);
				}
			}
			else
			{
				errorreq(NULL, "Not enough memory.", NULL);
			}
		}	
	}
}


int rbf_handle_password(STRPTR password)
{
	char buf[16];
	int r;
	
	if (readn(sock, buf, 16) == 16)
	{         
		EncryptPassword(buf, password);

		if (sendn(sock, buf, 16) == 16)
		{
			if (readn(sock, &r, 4) == 4)
			{
				switch (r)
				{
					case 0: /* ok */
						return (TRUE);

					case 1: /* failed */
						errorreq(NULL, "Wrong password.", "Quit");
						break;

					case 2: /* too many attempts */
						errorreq(NULL, "Too many attempts.", "Quit");
						break;

					default:
						errorreq(NULL, "Server is playing up.", "Quit");
						break;
				}
			}
		}
	}
	return (FALSE);
}

int rbf_clientinit(int shared)
{
	struct screeninfo si;
	int r;
	short int t = shared;

	if (sendn(sock, &t, 1))
	{
		if (readn(sock, &si, 20) == 20)
		{
			screen_width = (ULONG)si.width;
			screen_height = (ULONG)si.height;

			DB(("width == %ld, height == %ld\n", screen_width, screen_height));

			if (readn(sock, &r, 4) == 4)
			{
				/* eat up the useless screenname (TOFIX: use it later maybe ?) */
				if (r < 4096)
				{
					STRPTR p;
					if (p = malloc(r + 1))
					{
						DB(("success !\n"));
						free(p);
						return (TRUE);
					}
				}
			}
		}
	}
	return (FALSE);
}

int rbf_init(STRPTR host, int port, STRPTR password)
{
	DB(("initializing..\n"));
	if (!host)
	{
		errorreq(NULL, "No HOST supplied.", NULL);
		return (FALSE);
	}

	if (port < 1 || port > 65535)
	{
		errorreq(NULL, "Wrong PORT value.", NULL);
		return (FALSE);
	}

	if (tcp_connect(host, port))
	{
		DB(("connection ok\n"));
		if (rbf_version(3, 3))
		{
			DB(("got version info\n"));
			if (rbf_auth(password))
			{
				DB(("got authentication\n"));
				if (rbf_clientinit(TRUE))
				{
					DB(("client initialized properly\n"));
					me.msgtype = 5;
					me.x = screen_width / 2;
					me.y = screen_height / 2;
					ke.msgtype = 4;

					return (TRUE);
				}
			}
		}
	}
	return (FALSE);
}


void rbf_cleanup()
{
	tcp_disconnect();
}

struct convkey {
	ULONG code;
	ULONG ksd;
};

static struct convkey convtable[] = {
	0x0f, 0xff63, /* insert keypad XXX */
	0x1d, 0xff57, /* end keypad XXX */
	0x1f, 0xff56, /* page down keypad XXX */
	0x3c, 0xffff, /* delete keypad XXX */
	0x3d, 0xff50, /* home keypad XXX */
	0x3f, 0xff55, /* page up keypad XXX */
	0x41, 0xff08, /* backspace */
	0x42, 0xff09, /* tab */
	0x43, 0xff0d, /* enter */
	0x44, 0xff0d, /* return */
	0x45, 0xff1b, /* escape */
	0x46, 0xffff, /* delete */
	0x4a, 0xffad, /* keypad: '-' */
	0x4b, 0xffc8, /* F11 */
	0x4c, 0xff52, /* up */
	0x4d, 0xff54, /* down */
	0x4e, 0xff53, /* right */
	0x4f, 0xff51, /* left */
	0x50, 0xffbe, /* F1 */
	0x51, 0xffbf, /* F2 */
	0x52, 0xffc0, /* F3 */
	0x53, 0xffc1, /* F4 */
	0x54, 0xffc2, /* F5 */
	0x55, 0xffc3, /* F6 */
	0x56, 0xffc4, /* F7 */
	0x57, 0xffc5, /* F8 */
	0x58, 0xffc6, /* F9 */
	0x59, 0xffc7, /* F10 */
	0x5c, 0xffaf, /* keypad: '/' */
	0x5d, 0xffaa, /* keypad: '*' */
	0x5e, 0xffab, /* keypad: '+' */
	0x60, 0xffe1, /* lshift */
	0x61, 0xffe2, /* rshift */
	0x63, 0xffe3, /* ctrl XXX there's no right ctrl on the amiga */
	0x64, 0xffe9, /* lalt */
	0x65, 0xffea, /* ralt */
	0x66, 0xffe7, /* lamiga */
	0x67, 0xffe8, /* ramiga */
	0x6f, 0xffc9, /* F12 */
	NULL, NULL
};

struct deadkey {
	char code;
	char sequence[2];
};

/*
 * The following table is required for
 * proper deadkey convertion. Unfortunately
 * it seems that the server will make some
 * assumptions about deadkeys. Eg. '' works but
 * '' doesn't..
 */
static struct deadkey deadkeytable[] = {
	'', '`', 'A',
	'', '\'', 'A',
	'', '^', 'A',
	'', '~', 'A',
	'', '', 'A',
	'', '`', 'E',
	'', '\'', 'E',
	'', '^', 'E',
	'', '', 'E',
	'', '`', 'I',
	'', '\'', 'I',
	'', '^', 'I',
	'', '', 'I',
	'', '`', 'O',
	'', '\'', 'O',
	'', '^', 'O',
	'', '~', 'O',
	'', '', 'O',
	'', '`', 'U',
	'', '\'', 'U',
	'', '^', 'U',
	'', '', 'U',
	'', '\'', 'Y',
	'', '`', 'a',
	'', '\'', 'a',
	'', '^', 'a',
	'', '~', 'a',
	'', '', 'a',
	'', '`', 'e',
	'', '\'', 'e',
	'', '^', 'e',
	'', '', 'e',
	'', '`', 'i',
	'', '\'', 'i',
	'', '^', 'i',
	'', '', 'i',
	'', '~', 'n',
	'', '`', 'o',
	'', '\'', 'o',
	'', '^', 'o',
	'', '~', 'o',
	'', '', 'o',
	'', '`', 'u',
	'', '\'', 'u',
	'', '^', 'u',
	'', '', 'u',
	'', '\'', 'y',
	'', '', 'y',
	'\'', '\'', ' ',
	'`', '`', ' ',
	'^', '^', ' ',
	'', '', ' ',
	'~', '~', ' ',
	NULL, NULL, NULL
};

static int was_dead;
static int eatup;

//void dprintf(char *, ...);

int	rbf_send_keyevent(struct InputEvent *ie)
{
	ULONG i = 0;
	int mapped = 0;

	if (ie->ie_Code & IECODE_UP_PREFIX)
	{
		ke.down = 0;
		ie->ie_Code &= ~IECODE_UP_PREFIX;
	}
	else
	{
		ke.down = 1;
	}
	
	#if 0
	/*
	 * We exit when the user presses the Help key
	 */
	if (ie->ie_Code == 0x5f)
	{
		DB(("quitting..\n"));
		return (RBF_QUIT);
	}
	#endif

	if (eatup)
	{
		eatup = FALSE;
		return (RBF_OK);
	}

	/*
	 * Speed up. Keep it in sync with
	 * the table.
	 *
	 */
	if ((ie->ie_Code >= 0x3c && ie->ie_Code <= 0x6f) ||
		ie->ie_Code == 0x0f ||
		ie->ie_Code == 0x1d ||
		ie->ie_Code == 0x1f)
	{
	    while (convtable[i].code)
		{
			if (ie->ie_Code == convtable[i].code)
			{
				ke.key = convtable[i].ksd;
				DB(("0x%lx mapped, key: 0x%lx, %s\n", convtable[i].code, ke.key, ke.down ? "down" : "up"));
				mapped = TRUE;
				break;
			}
			i++;
		}
	}

	if (!mapped)
	{
		struct InputEvent mie;
		char k;

		mie.ie_Class = IECLASS_RAWKEY;
		mie.ie_SubClass = ie->ie_SubClass;
		mie.ie_Code = ie->ie_Code;
		mie.ie_Qualifier = ie->ie_Qualifier & ~(IEQUALIFIER_CONTROL); /* no CTRL, we need the real key. server already knows CTRL was pressed. XXX: what about other keys? */
		mie.ie_EventAddress = ie->ie_EventAddress;

		//dprintf("key event: 0x%lx\n", ie->ie_Code);

		if (MapRawKey(&mie, &k, 1, NULL) == 1)
		{
			DB(("key: 0x%lx, %s\n", (ULONG)k, ke.down ? "down" : "up"));
			ke.key = (ULONG)k;
		}
		else
		{
			DB(("unknown mapping key (%ld)!\n", ie->ie_Code));
			was_dead = TRUE;
			return (RBF_OK);
		}
	}

	/*
	 * Handle deadkeys.
	 */
	if (!args.no_deadkey && was_dead && !mapped) /*&& (ke.key > (ULONG)'' || ke.key == (ULONG)'' || ke.key == (ULONG)'^'))*/
	{
		i = 0;
		was_dead = FALSE;
		while (deadkeytable[i].code)
		{
			if (ke.key == (ULONG)deadkeytable[i].code)
			{
				DB(("sending deadkey %ld\n", (ULONG)deadkeytable[i].code));

				ke.down = 1;
				ke.key = (ULONG)deadkeytable[i].sequence[0];
				if (sendn(sock, &ke, sizeof(struct keyevent)) == sizeof(struct keyevent))
				{
					ke.down = 0;
					if (sendn(sock, &ke, sizeof(struct keyevent)) == sizeof(struct keyevent))
					{
						ke.down = 1;
						ke.key = (ULONG)deadkeytable[i].sequence[1];
						if (sendn(sock, &ke, sizeof(struct keyevent)) == sizeof(struct keyevent))
						{
							ke.down = 0;
							if (sendn(sock, &ke, sizeof(struct keyevent)) == sizeof(struct keyevent))
							{
								/*
								 * Eat the next deadkey up.
								 */
								eatup = TRUE;

								return (RBF_OK);
							}
						}
					}
				}
				return (RBF_BROKENLINK);
			}
			i++;
		}
	}

	if (sendn(sock, &ke, sizeof(struct keyevent)) == sizeof(struct keyevent))
	{
		return (RBF_OK);
	}
	else
	{
		return (RBF_BROKENLINK);
	}
}


int	rbf_send_mouseevent(struct InputEvent *ie)
{
	switch (ie->ie_Code)
	{
		case 0xe8:
			me.butmask &= ~1;
			break;

		case IECODE_LBUTTON:
			me.butmask |= 1;
			break;

		case 0xe9:
			me.butmask &= ~4;
			break;

		case IECODE_RBUTTON:
			me.butmask |= 4;
			break;

		case 0xea:
			me.butmask &= ~2;
			break;

		case IECODE_MBUTTON:
			me.butmask |= 2;
			break;

		case NM_WHEEL_UP:
			me.butmask |= 8;
			break;

		case NM_WHEEL_DOWN:
			me.butmask |= 16;
			break;
	}
	
	me.x += ie->ie_X;
	if (me.x < 0)
	{
		me.x = 0;
	}
	else if (me.x > screen_width)
	{
		me.x = screen_width;
	}

	me.y += ie->ie_Y;
	if (me.y < 0)
	{
		me.y = 0;
	}
	else if (me.y > screen_height)
	{
		me.y = screen_height;
	}

	if (sendn(sock, &me, sizeof(struct mouseevent)) != sizeof(struct mouseevent))
	{
		return (RBF_BROKENLINK);
	}

	/*
	 * If that was a mouse wheel event, we need
	 * to send another one with the button cleared
	 * or rbfserver for x11 would be confused.
	 */
	if (me.butmask & (8 | 16))
	{
		me.butmask &= ~(8 | 16);

		if (sendn(sock, &me, sizeof(struct mouseevent)) != sizeof(struct mouseevent))
		{
			return (RBF_BROKENLINK);
		}
	}
	return (RBF_OK);
}
